/***************************************************************************
 *   Copyright (C) 2010 by Naomasa Matsubayashi   *
 *   picosynth@quaternion.sakura.ne.jp   *
 *                                                                         *
 *   All rights reserved.                                                  *
 *                                                                         *
 * Redistribution and use in source and binary forms, with or without      *
 * modification, are permitted provided that the following conditions are  *
 * met:                                                                    *
 *                                                                         *
 *  1. Redistributions of source code must retain the above copyright      *
 *     notice, this list of conditions and the following disclaimer.       *
 *  2. Redistributions in binary form must reproduce the above copyright   *
 *     notice, this list of conditions and the following disclaimer in the *
 *     documentation and/or other materials provided with the distribution.*
 *                                                                         *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS     *
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT       *
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT      *
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,   *
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT        *
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,   *
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY   *
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT     *
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE   *
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.    *
 ***************************************************************************/

#ifndef PICOSYNTH_CLOCK_H
#define PICOSYNTH_CLOCK_H

#include <stdint.h>
#include <types.h>
#include <fixed.h>

#define TONE_CLOCK0 0
#define TONE_CLOCK1 1
#define LOCAL_CLOCK 2

/*
 * クロック
 * 現在時刻を管理する構造体
 * 時刻には以下の3つの種類がある
 * グローバルクロック	演奏開始からの経過時間
 * ローカルクロック	現在処理しているキーが押されてからの経過時間の定数倍(エンベロープ用)
 * トーンクロック		現在処理しているキーが押されてからの経過時間の定数倍(トーン用)
 * トーンクロックはデチューン用に2つ備わっている
 * グローバルクロックはpicosynthではどこでも使用していないので参照する
 * ための関数は用意していない
 */

typedef struct {
  Time global_tick;
  ToneTime current_tone_time[ 3 ];
  ToneTime tone_tick[ 3 ];
  Bool local_clock_overflow;
} Clock;

// クロックを初期化する
INLINE void clkInit( Clock *_clock ) {
  _clock->global_tick = 0x10000ul / FREQUENCY;
}

// クロックを1サンプル分進める
INLINE void clkTick( Clock *_clock ) {
  _clock->current_tone_time[ TONE_CLOCK0 ] += _clock->tone_tick[ TONE_CLOCK0 ];
  _clock->current_tone_time[ TONE_CLOCK1 ] += _clock->tone_tick[ TONE_CLOCK1 ];
  ToneTime old_clock = _clock->current_tone_time[ LOCAL_CLOCK ];
  _clock->current_tone_time[ LOCAL_CLOCK ] += _clock->tone_tick[ LOCAL_CLOCK ];
  if( old_clock > _clock->current_tone_time[ LOCAL_CLOCK ] )
    _clock->local_clock_overflow = TRUE;
}

// ローカルクロックを設定する
INLINE void clkSetLocal( Clock *_clock, uint8_t _freq ) {
  _clock->current_tone_time[ LOCAL_CLOCK ] = 0;
  _clock->tone_tick[ LOCAL_CLOCK ] = ( _clock->global_tick * _freq ) >> 4;
  _clock->local_clock_overflow = FALSE;
}

// ローカルクロックをリセットする
INLINE void clkResetLocal( Clock *_clock ) {
  _clock->current_tone_time[ LOCAL_CLOCK ] = 0;
  _clock->local_clock_overflow = FALSE;
}

// トーンクロックを設定する
INLINE void clkSetTone( Clock *_clock, Freq _freq, ToneTime _detune0, ToneTime _detune1 ) {
  _clock->current_tone_time[ TONE_CLOCK0 ] = 0;
  _clock->tone_tick[ TONE_CLOCK0 ] = ( _clock->global_tick * ( _freq + _detune0 ) ) >> 16;
  _clock->current_tone_time[ TONE_CLOCK1 ] = 0;
  _clock->tone_tick[ TONE_CLOCK1 ] = ( _clock->global_tick * ( _freq + _detune1 ) ) >> 16;
}

// ローカルクロックの値を得る
INLINE Fixed8 clkGetLocalTime( Clock *_clock ) {
  return _clock->current_tone_time[ LOCAL_CLOCK ] >> 9;
}

// ローカルクロックが最後にリセットされてからオーバーフローしていたらTRUEを返す
INLINE Bool clkLocalClockOverflowed( Clock *_clock ) {
  return _clock->local_clock_overflow;
}

// トーンクロックの値を得る
INLINE Fixed8 clkGetToneTime( Clock *_clock, uint8_t _select ) {
  return _clock->current_tone_time[ _select ] >> 9;
}

#endif
