/***************************************************************************
 *   Copyright (C) 2010 by Naomasa Matsubayashi   *
 *   picosynth@quaternion.sakura.ne.jp   *
 *                                                                         *
 *   All rights reserved.                                                  *
 *                                                                         *
 * Redistribution and use in source and binary forms, with or without      *
 * modification, are permitted provided that the following conditions are  *
 * met:                                                                    *
 *                                                                         *
 *  1. Redistributions of source code must retain the above copyright      *
 *     notice, this list of conditions and the following disclaimer.       *
 *  2. Redistributions in binary form must reproduce the above copyright   *
 *     notice, this list of conditions and the following disclaimer in the *
 *     documentation and/or other materials provided with the distribution.*
 *                                                                         *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS     *
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT       *
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT      *
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,   *
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT        *
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,   *
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY   *
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT     *
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE   *
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.    *
 ***************************************************************************/

#ifndef PICOSYNTH_DRUM_H
#define PICOSYNTH_DRUM_H

#include <stdint.h>
#include <types.h>
#include <fixed.h>
#include <module.h>
#include <monophony.h>
#include <instrument.h>

/*
 * ドラム
 * ドラムはモジュールに長さと音階を追加したものである
 * ドラム演奏時は楽譜から与えられた長さと音階は無視され、ここで指定された長さと音階で
 * 演奏される
 */
typedef struct {
  Module module;
  uint8_t length;
  NoteNo note_number;
} Drum;

// ドラムの音階を得る
INLINE NoteNo drmGetNoteNumber ( const Drum *_drum ) {
  return _drum->note_number;
}

// ドラムの長さを得る
INLINE uint8_t drmGetLength ( const Drum *_drum ) {
  return _drum->length;
}

// ドラムのモジュールを得る
INLINE const Module *drmGetModule ( const Drum *_drum ) {
  return & ( _drum->module );
}

/*
 * ドラムキット
 * ドラムキットはドラムテーブルを参照し、指定された音階に対応するドラムを演奏する
 * 演奏対象がドラムである以外は基本的にインスツルメントと同じ振る舞いをする
 */
typedef struct {
  const Drum **drum;
  Monophony monophony[ POLYPHONY_NUM ];
} DrumKit;

// ドラムキットを初期化する
INLINE void dktInit ( DrumKit *_drumkit, const Drum **_drum ) {
  _drumkit->drum = _drum;
  uint8_t index;
  for ( index = 0; index != POLYPHONY_NUM; index++ )
    mpyInit ( _drumkit->monophony + index );
}

// モノフォニーを得る
INLINE Monophony *dktGetMonophony ( DrumKit *_drumkit, uint8_t _index ) {
  return _drumkit->monophony + _index;
}

// ドラムを得る
INLINE const Drum *dktGetDrum ( DrumKit *_drumkit, NoteNo _index ) {
  return _drumkit->drum[ _index ];
}

// 音階に対応するドラムが存在するかどうか調べる
INLINE Bool dktDrumIsDefined ( DrumKit *_drumkit, NoteNo _note_number ) {
  return dktGetDrum ( _drumkit, _note_number ) != NULL;
}

// ノートオン内部呼び出し用関数
INLINE void dktNoteOnInternal ( DrumKit *_drumkit, NoteNo _note_number, uint16_t _current_time, uint8_t _index ) {
  Monophony *monophony = dktGetMonophony ( _drumkit, _index );
  const Drum *drum = dktGetDrum ( _drumkit, _note_number );
  nteOn ( mpyGetNote ( monophony ), drmGetNoteNumber ( drum ) );
  clkSetLocal ( mpyGetClock ( monophony ), evlGetFreq ( mdrGetEnvelope ( drmGetModule( drum ) ) ) );
  clkSetTone ( mpyGetClock ( monophony ), nteGetFreq ( mpyGetNote ( monophony ) ),
               wavGetDetune ( mdrGetWave ( drmGetModule ( drum ), 0 ) ),
               wavGetDetune ( mdrGetWave ( drmGetModule ( drum ), 1 ) ) );
  nteOn ( mpyGetNote ( monophony ), _note_number );
  mpySetNoteOffTime ( monophony, _current_time + drmGetLength ( drum ) );
}

// ノートオフ
INLINE void dktNoteOff ( DrumKit *_drumkit, uint16_t _current_time ) {
  uint8_t index;
  for ( index = 0; index != POLYPHONY_NUM; index++ ) {
    if ( nteGetState ( mpyGetNote ( dktGetMonophony ( _drumkit, index ) ) ) == NOTE_ON ) {
      if ( mpyGetNoteOffTime ( dktGetMonophony ( _drumkit, index ) ) <= _current_time ) {
        nteOff ( mpyGetNote ( dktGetMonophony ( _drumkit, index ) ) );
        clkResetLocal ( mpyGetClock ( dktGetMonophony ( _drumkit, index ) ) );
      }
    }
  }
}

// ノートオン
INLINE void dktNoteOn ( DrumKit *_drumkit, NoteNo _note_number, uint16_t _current_time ) {
  if ( dktDrumIsDefined ( _drumkit, _note_number ) ) {
    uint8_t index;
    for ( index = 0; index != POLYPHONY_NUM; index++ ) {
      if ( nteGetState ( mpyGetNote ( dktGetMonophony ( _drumkit, index ) ) ) == NOTE_IDLE ) {
        dktNoteOnInternal ( _drumkit, _note_number, _current_time, index );
        return;
      }
    }
    uint16_t oldest_note_off_time = 0xFFFF;
    uint8_t oldest_index = POLYPHONY_NUM;
    for ( index = 0; index != POLYPHONY_NUM; index++ ) {
      Monophony *monophony = dktGetMonophony ( _drumkit, index );
      if ( nteGetState ( mpyGetNote ( monophony ) ) == NOTE_OFF ) {
        uint16_t current_note_off_time = mpyGetNoteOffTime ( monophony );
        if ( current_note_off_time <= oldest_note_off_time ) {
          oldest_note_off_time = current_note_off_time;
          oldest_index = index;
        }
      }
    }
    if ( oldest_index != POLYPHONY_NUM ) {
      dktNoteOnInternal ( _drumkit, _note_number, _current_time, oldest_index );
      return;
    }
    oldest_note_off_time = 0xFFFF;
    oldest_index = POLYPHONY_NUM;
    for ( index = 0; index != POLYPHONY_NUM; index++ ) {
      Monophony *monophony = dktGetMonophony ( _drumkit, index );
      uint16_t current_note_off_time = mpyGetNoteOffTime ( monophony );
      if ( current_note_off_time <= oldest_note_off_time ) {
        oldest_note_off_time = current_note_off_time;
        oldest_index = index;
      }
    }
    nteOff ( mpyGetNote ( dktGetMonophony ( _drumkit, oldest_index ) ) );
    clkResetLocal ( mpyGetClock ( dktGetMonophony ( _drumkit, oldest_index ) ) );
    dktNoteOnInternal ( _drumkit, _note_number, _current_time, oldest_index );
    return;
  }
}

// 現在の時刻のサンプルを得る
INLINE Fixed8_8 dktExec ( DrumKit *_drumkit ) {
  uint8_t index;
  Fixed8_8 sum = 0;
  for ( index = 0; index != POLYPHONY_NUM; index++ ) {
    Monophony *monophony = dktGetMonophony ( _drumkit, index );
    if ( nteGetState ( mpyGetNote ( monophony ) ) != NOTE_IDLE ) {
      const Drum *drum = dktGetDrum ( _drumkit, nteGetNoteNumber ( mpyGetNote ( monophony ) ) );
      sum += mdrExec ( drmGetModule ( drum ), mpyGetClock ( monophony ), mpyGetNote ( monophony ) );
      clkTick ( mpyGetClock ( monophony ) );
    }
  }
  return sum;
}


#endif
