/***************************************************************************
 *   Copyright (C) 2010 by Naomasa Matsubayashi   *
 *   picosynth@quaternion.sakura.ne.jp   *
 *                                                                         *
 *   All rights reserved.                                                  *
 *                                                                         *
 * Redistribution and use in source and binary forms, with or without      *
 * modification, are permitted provided that the following conditions are  *
 * met:                                                                    *
 *                                                                         *
 *  1. Redistributions of source code must retain the above copyright      *
 *     notice, this list of conditions and the following disclaimer.       *
 *  2. Redistributions in binary form must reproduce the above copyright   *
 *     notice, this list of conditions and the following disclaimer in the *
 *     documentation and/or other materials provided with the distribution.*
 *                                                                         *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS     *
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT       *
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT      *
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,   *
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT        *
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,   *
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY   *
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT     *
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE   *
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.    *
 ***************************************************************************/

#ifndef PICOSYNTH_ENVELOPE_H
#define PICOSYNTH_ENVELOPE_H

#include <stdint.h>
#include <types.h>
#include <fixed.h>

#define RELEASE_SPEED( speed ) \
  ( (uint16_t)( ( ( (uint32_t)speed << 14 ) / FREQUENCY ) ) )

/*
 * エンベロープ
 * 音の立ち上がりや伸びをコントロールする
 * picosynthでは実装を簡素にするためにアタックからディケイまでは波形テーブルを使い、
 * リリースだけ真面目に計算する
 */
typedef struct {
  const Fixed8 *table;
  uint16_t release_speed;
  uint16_t table_size;
  uint8_t table_size_pot;
  uint8_t freq;
} Envelope;

// エンベロープの波形テーブルの再生周波数を得る
INLINE uint8_t evlGetFreq( const Envelope *_envelope ) {
  return _envelope->freq;
}

// 現在のエンベロープの値を得る
INLINE Fixed8 evlExec( const Envelope *_envelope, Clock *_clock, Note *_note ) {
  Fixed8 current_time = clkGetLocalTime( _clock );
  if( nteGetState( _note ) == NOTE_ON ) {
    if( nteIsSustain( _note ) ) {
      Fixed8 envelope_level = _envelope->table[ _envelope->table_size - 1 ];
      nteSetReleaseLevel( _note, envelope_level );
      return envelope_level;
    }
    else if( clkLocalClockOverflowed( _clock ) ) {
      nteSetSustain( _note );
      Fixed8 envelope_level = _envelope->table[ _envelope->table_size - 1 ];
      nteSetReleaseLevel( _note, envelope_level );
      return envelope_level;
    }
    else {
      Fixed8 envelope_level = _envelope->table[ current_time >> ( 7 - _envelope->table_size_pot ) ];
      nteSetReleaseLevel( _note, envelope_level );
      return envelope_level;
    }
  }
  else {
    nteDecrementReleaseLevel( _note, _envelope->release_speed );
    Fixed8 temp = nteGetReleaseLevel( _note );
    if( temp <= 0 ) {
      nteSetEnd( _note );
      return 0;
    }
    else
      return temp;
  }
}

#endif
