/***************************************************************************
 *   Copyright (C) 2010 by Naomasa Matsubayashi   *
 *   picosynth@quaternion.sakura.ne.jp   *
 *                                                                         *
 *   All rights reserved.                                                  *
 *                                                                         *
 * Redistribution and use in source and binary forms, with or without      *
 * modification, are permitted provided that the following conditions are  *
 * met:                                                                    *
 *                                                                         *
 *  1. Redistributions of source code must retain the above copyright      *
 *     notice, this list of conditions and the following disclaimer.       *
 *  2. Redistributions in binary form must reproduce the above copyright   *
 *     notice, this list of conditions and the following disclaimer in the *
 *     documentation and/or other materials provided with the distribution.*
 *                                                                         *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS     *
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT       *
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT      *
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,   *
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT        *
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,   *
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY   *
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT     *
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE   *
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.    *
 ***************************************************************************/

#ifndef PICOSYNTH_INSTRUMENT_H
#define PICOSYNTH_INSTRUMENT_H

#include <stdint.h>
#include <types.h>
#include <fixed.h>
#include <module.h>
#include <monophony.h>

#define POLYPHONY_NUM 4

/*
 * 楽器
 * POLYPHONY_NUMで指定される数のモノフォニーを保持し、ノートオン時にどのモノフォニーに
 * その処理を割り当てるかを決定する
 */
typedef struct {
  const Module *module;
  Monophony monophony[ POLYPHONY_NUM ];
} Instrument;

// インスツルメントを初期化する
INLINE void istInit( Instrument *_instrument, const Module *_module ) {
  _instrument->module = _module;
  uint8_t index;
  for( index = 0; index != POLYPHONY_NUM; index++ )
    mpyInit( _instrument->monophony + index );
}

// モノフォニーを得る
INLINE Monophony *istGetMonophony( Instrument *_instrument, uint8_t _index ) {
  return _instrument->monophony + _index;
}

// モジュールを得る
INLINE const Module *istGetModule( Instrument *_instrument ) {
  return _instrument->module;
}

// ノートオン内部呼び出し用
INLINE void istNoteOnInternal( Instrument *_instrument, NoteNo _note_number, uint16_t _note_off_time, uint8_t _index ) {
  Monophony *monophony = istGetMonophony( _instrument, _index );
  nteOn( mpyGetNote( monophony ), _note_number );
  clkSetLocal( mpyGetClock( monophony ), evlGetFreq( mdrGetEnvelope( istGetModule( _instrument ) ) ) );
  clkSetTone( mpyGetClock( monophony ), nteGetFreq( mpyGetNote( monophony ) ),
              wavGetDetune( mdrGetWave( istGetModule( _instrument ), 0 ) ),
                            wavGetDetune( mdrGetWave( istGetModule( _instrument ), 1 ) ) );
  mpySetNoteOffTime( monophony, _note_off_time );
}

// ノートオフ
INLINE void istNoteOff( Instrument *_instrument, uint16_t _current_time ) {
  uint8_t index;
  for( index = 0; index != POLYPHONY_NUM; index++ ) {
    if( nteGetState( mpyGetNote( istGetMonophony( _instrument, index ) ) ) == NOTE_ON ) {
      if( mpyGetNoteOffTime( istGetMonophony( _instrument, index ) ) <= _current_time ) {
        nteOff( mpyGetNote( istGetMonophony( _instrument, index ) ) );
        clkResetLocal( mpyGetClock( istGetMonophony( _instrument, index ) ) );
      }
    }
  }
}

// ノートオン
INLINE void istNoteOn( Instrument *_instrument, NoteNo _note_number, uint16_t _note_off_time ) {
  uint8_t index;
  for( index = 0; index != POLYPHONY_NUM; index++ ) {
    if( nteGetState( mpyGetNote( istGetMonophony( _instrument, index ) ) ) == NOTE_IDLE ) {
      istNoteOnInternal( _instrument, _note_number, _note_off_time, index );
      return;
    }
  }
  uint16_t oldest_note_off_time = 0xFFFF;
  uint8_t oldest_index = POLYPHONY_NUM;
  for( index = 0; index != POLYPHONY_NUM; index++ ) {
    Monophony *monophony = istGetMonophony( _instrument, index );
    if( nteGetState( mpyGetNote( monophony ) ) == NOTE_OFF ) {
      uint16_t current_note_off_time = mpyGetNoteOffTime( monophony );
      if( current_note_off_time <= oldest_note_off_time ) {
        oldest_note_off_time = current_note_off_time;
        oldest_index = index;
      }
    }
  }
  if( oldest_index != POLYPHONY_NUM ) {
    istNoteOnInternal( _instrument, _note_number, _note_off_time, oldest_index );
    return;
  }
  oldest_note_off_time = 0xFFFF;
  oldest_index = POLYPHONY_NUM;
  for( index = 0; index != POLYPHONY_NUM; index++ ) {
    Monophony *monophony = istGetMonophony( _instrument, index );
    uint16_t current_note_off_time = mpyGetNoteOffTime( monophony );
    if( current_note_off_time <= oldest_note_off_time ) {
      oldest_note_off_time = current_note_off_time;
      oldest_index = index;
    }
  }
  nteOff( mpyGetNote( istGetMonophony( _instrument, oldest_index ) ) );
  clkResetLocal( mpyGetClock( istGetMonophony( _instrument, oldest_index ) ) );
  istNoteOnInternal( _instrument, _note_number, _note_off_time, oldest_index );
  return;
}

// 現在のサンプルを得る
INLINE Fixed8_8 istExec( Instrument *_instrument ) {
  uint8_t index;
  Fixed8_8 sum = 0;
  for( index = 0; index != POLYPHONY_NUM; index++ ) {
    Monophony *monophony = istGetMonophony( _instrument, index );
    if( nteGetState( mpyGetNote( monophony ) ) != NOTE_IDLE ) {
      sum += mdrExec( istGetModule( _instrument ), mpyGetClock( monophony ), mpyGetNote( monophony ) );
      clkTick( mpyGetClock( monophony ) );
    }
  }
  return sum;
}
#endif
