/***************************************************************************
 *   Copyright (C) 2010 by Naomasa Matsubayashi   *
 *   picosynth@quaternion.sakura.ne.jp   *
 *                                                                         *
 *   All rights reserved.                                                  *
 *                                                                         *
 * Redistribution and use in source and binary forms, with or without      *
 * modification, are permitted provided that the following conditions are  *
 * met:                                                                    *
 *                                                                         *
 *  1. Redistributions of source code must retain the above copyright      *
 *     notice, this list of conditions and the following disclaimer.       *
 *  2. Redistributions in binary form must reproduce the above copyright   *
 *     notice, this list of conditions and the following disclaimer in the *
 *     documentation and/or other materials provided with the distribution.*
 *                                                                         *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS     *
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT       *
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT      *
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,   *
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT        *
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,   *
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY   *
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT     *
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE   *
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.    *
 ***************************************************************************/

#ifndef PICOSYNTH_NOTE_H
#define PICOSYNTH_NOTE_H

#include <stdint.h>
#include <types.h>
#include <fixed.h>

typedef unsigned char NoteNo;
typedef unsigned char NoteState;
#define NOTE_ON ((NoteState)1)
#define NOTE_OFF ((NoteState)2)
#define NOTE_IDLE ((NoteState)0)

/*
 * 1オクターブ分の周波数テーブル
 */
const Freq freq_table[ 12 ] = {
  274334289, 290647054, 307929828, 326240288, 345639545, 366192342, 387967272, 411037006, 435478539, 461373440, 488808132, 517874176,
};

/*
 * 音階
 * 現在処理しているキーの音階と状態を保持するステートマシン
 */
typedef struct {
  NoteNo note_number;
  NoteState note_state;
  Bool is_end;
  Bool is_sustain;
  Fixed16 last_level;
} Note;

// 音階を初期化する
INLINE void nteInit( Note *_note ) {
  _note->note_state = NOTE_IDLE;
}

// 音階の現在の状態を得る
INLINE NoteState nteGetState( Note *_note ) {
  return _note->note_state;
}

// 音階の現在の状態をノートオン(キーが押された)にする
INLINE void nteOn( Note *_note, NoteNo _note_number ) {
  _note->note_number = _note_number;
  _note->note_state = NOTE_ON;
  _note->is_end = FALSE;
  _note->is_sustain = FALSE;
}

// 音階の現在の状態をノートオフ(キーが離された)にする
INLINE void nteOff( Note *_note ) {
  _note->note_state = NOTE_OFF;
  _note->is_sustain = FALSE;
  _note->last_level <<= 8;
}

// 音階の現在の状態をアイドル(処理する必要がない)にする
INLINE void nteSetEnd( Note *_note ) {
  _note->note_state = NOTE_IDLE;
}

// 音階を得る
INLINE NoteNo nteGetNoteNumber( Note *_note ) {
  return _note->note_number;
}

// 周波数を得る
INLINE Freq nteGetFreq( Note *_note ) {
  Freq temp = freq_table[ _note->note_number % 12 ];
  if( _note->note_number >= 0x78 ) {
    temp <<= 1;
    return temp;
  }
  else {
    temp >>= ( 9 - ( _note->note_number / 12 ) );
    return temp;
  }
}

// 次の瞬間ノートオフした場合のエンベロープのリリースの開始値を設定する
INLINE void nteSetReleaseLevel( Note *_note, Fixed8 _level ) {
  _note->last_level = _level;
  if( _note->last_level <= 0 ) {
    nteSetEnd( _note );
    _note->last_level = 0;
  }
}

// エンベロープのリリースの現在値を得る
INLINE Fixed8 nteGetReleaseLevel( Note *_note ) {
  return _note->last_level >> 8;
}

// エンベロープのリリースの現在値を指定された量減らす
INLINE void nteDecrementReleaseLevel( Note *_note, uint16_t _grad ) {
  _note->last_level -= _grad;
  if( _note->last_level <= 0 ) {
    nteSetEnd( _note );
    _note->last_level = 0;
  }
}

// エンベロープが現在サスティンならTRUEを返す
INLINE Bool nteIsSustain( Note *_note ) {
  return _note->is_sustain;
}

// サスティンフラグを立てる
INLINE void nteSetSustain( Note *_note ) {
  _note->is_sustain = TRUE;
}

#endif
