/***************************************************************************
 *   Copyright (C) 2010 by Naomasa Matsubayashi   *
 *   picosynth@quaternion.sakura.ne.jp   *
 *                                                                         *
 *   All rights reserved.                                                  *
 *                                                                         *
 * Redistribution and use in source and binary forms, with or without      *
 * modification, are permitted provided that the following conditions are  *
 * met:                                                                    *
 *                                                                         *
 *  1. Redistributions of source code must retain the above copyright      *
 *     notice, this list of conditions and the following disclaimer.       *
 *  2. Redistributions in binary form must reproduce the above copyright   *
 *     notice, this list of conditions and the following disclaimer in the *
 *     documentation and/or other materials provided with the distribution.*
 *                                                                         *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS     *
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT       *
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT      *
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,   *
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT        *
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,   *
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY   *
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT     *
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE   *
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.    *
 ***************************************************************************/

#ifndef PICOSYNTH_PLAYERRUNTIMEDATA_H
#define PICOSYNTH_PLAYERRUNTIMEDATA_H

#include <stdint.h>
#include <fixed.h>
#include <truckruntimedata.h>
#include <normalize.h>
#include <truck.h>
#include <score.h>
#include <drumtruckruntimedata.h>

/*
 * プレイヤー
 * 楽譜と現在時刻を保持する
 */
typedef struct {
  const Score *score;
  uint16_t sample_per_read;
} Player;

/*
 * プレイヤー実行時データ
 * 再生時に必要になる情報をまとめたもの
 */
typedef struct {
  Normalize normalize;
  PCM pcm;
  TruckRuntimeData trd[ TRUCK_NUM ];
  DrumTruckRuntimeData dtrd;
  uint16_t current_position;
} PlayerRuntimeData;

// プレイヤー実行時データを初期化
INLINE void prdInit( PlayerRuntimeData *_prd, const Player *_player ) {
  nmlInit( &( _prd->normalize ) );
  pcmOpen( &( _prd->pcm ) );
  uint8_t truck_index;
  for( truck_index = 0; truck_index != TRUCK_NUM; truck_index++ )
    trdInit( _prd->trd + truck_index, _player->score->truck + truck_index );
  _prd->current_position = 0;
  dtrInit( &( _prd->dtrd ), &( _player->score->drum_truck ) );
}

// プレイヤー実行時データを破棄
INLINE void prdDestroy( PlayerRuntimeData *_prd ) {
  pcmClose( &( _prd->pcm ) );
}

// トラックを取得
INLINE TruckRuntimeData *prdGetTruck( PlayerRuntimeData *_prd, uint8_t _index ) {
  return _prd->trd + _index;
}

// ドラムとラックを取得
INLINE DrumTruckRuntimeData *prdGetDrumTruck( PlayerRuntimeData *_prd ) {
  return &( _prd->dtrd );
}

// 正規化を取得
INLINE Normalize *prdGetNormalize( PlayerRuntimeData *_prd ) {
  return &( _prd->normalize );
}

// 楽譜が終端に達していたらTRUEを返す
INLINE Bool prdIsEnd( PlayerRuntimeData *_prd ) {
  Bool is_end = FALSE;
  uint8_t truck_index;
  for( truck_index = 0; truck_index != TRUCK_NUM; truck_index++ )
    if( trdIsAvailable( _prd->trd + truck_index ) )
      is_end |= trdIsEnd( _prd->trd + truck_index, _prd->current_position );
  if( dtrIsAvailable( &( _prd->dtrd ) ) )
    is_end |= dtrIsEnd( &( _prd->dtrd ), _prd->current_position );
  return is_end;
}

// 時刻を進める
INLINE void prdAdvance( PlayerRuntimeData *_prd ) {
  _prd->current_position++;
}

// 現在の時刻でのノートオン/ノートオフイベントを処理する
INLINE void prdControlNote( PlayerRuntimeData *_prd ) {
  uint8_t truck_index;
  for( truck_index = 0; truck_index != TRUCK_NUM; truck_index++ )
    if( trdIsAvailable( _prd->trd + truck_index ) )
      trdControlNote( _prd->trd + truck_index, _prd->current_position );
  if( dtrIsAvailable( &( _prd->dtrd ) ) )
    dtrControlNote( &( _prd->dtrd ), _prd->current_position );
}

// 現在の時刻での値を得る
INLINE void prdCalc( PlayerRuntimeData *_prd ) {
  uint8_t truck_index;
  Fixed8_8 sum = 0;
  for( truck_index = 0; truck_index != TRUCK_NUM; truck_index++ )
    if( trdIsAvailable( _prd->trd + truck_index ) )
      sum += trdCalc( _prd->trd + truck_index );
  if( dtrIsAvailable( &( _prd->dtrd ) ) )
    sum += dtrCalc( &( _prd->dtrd ) ) >> 1;
  sum = nmlExec( prdGetNormalize( _prd ), sum );
  pcmWrite( &( _prd->pcm ), sum );
}

#endif
